export default apkUtilsMethods;
export type InstallState = "unknown" | "notInstalled" | "newerVersionInstalled" | "sameVersionInstalled" | "olderVersionInstalled";
export type IsAppInstalledOptions = {
    /**
     * - The user id
     */
    user?: string | undefined;
};
export type StartUriOptions = {
    /**
     * - if `false` then adb won't wait
     * for the started activity to return the control
     */
    waitForLaunch?: boolean | undefined;
};
export type StartAppOptions = {
    /**
     * - The name of the application package
     */
    pkg: string;
    /**
     * - The name of the main application activity.
     * This or action is required in order to be able to launch an app.
     */
    activity?: string | undefined;
    /**
     * - The name of the intent action that will launch the required app.
     * This or activity is required in order to be able to launch an app.
     */
    action?: string | undefined;
    /**
     * - If this property is set to `true`
     * and the activity name does not start with '.' then the method
     * will try to add the missing dot and start the activity once more
     * if the first startup try fails.
     */
    retry?: boolean | undefined;
    /**
     * - Set it to `true` in order to forcefully
     * stop the activity if it is already running.
     */
    stopApp?: boolean | undefined;
    /**
     * - The name of the package to wait to on
     * startup (this only makes sense if this name is different from the one, which is set as `pkg`)
     */
    waitPkg?: string | undefined;
    /**
     * - The name of the activity to wait to on
     * startup (this only makes sense if this name is different from the one, which is set as `activity`)
     */
    waitActivity?: string | undefined;
    /**
     * - The number of milliseconds to wait until the
     * `waitActivity` is focused
     */
    waitDuration?: number | undefined;
    /**
     * - The number of the user profile to start
     * the given activity with. The default OS user profile (usually zero) is used
     * when this property is unset
     */
    user?: string | number | undefined;
    /**
     * - if `false` then adb won't wait
     * for the started activity to return the control
     */
    waitForLaunch?: boolean | undefined;
    category?: string | undefined;
    flags?: string | undefined;
    optionalIntentArguments?: string | undefined;
};
export type PackageActivityInfo = {
    /**
     * - The name of application package,
     *   for example 'com.acme.app'.
     */
    appPackage: string | null;
    /**
     * - The name of main application activity.
     */
    appActivity: string | null;
};
export type UninstallOptions = {
    /**
     * - The count of milliseconds to wait until the
     *          app is uninstalled.
     */
    timeout?: number | undefined;
    /**
     * - Set to true in order to keep the
     *          application data and cache folders after uninstall.
     */
    keepData?: boolean | undefined;
    /**
     * Whether to check if the app is installed prior to
     * uninstalling it. By default this is checked.
     */
    skipInstallCheck?: boolean | undefined;
};
export type CachingOptions = {
    /**
     * - The count of milliseconds to wait until the
     * app is uploaded to the remote location.
     */
    timeout?: number | undefined;
};
export type InstallOptions = {
    /**
     * - The count of milliseconds to wait until the
     *    app is installed.
     */
    timeout?: number | undefined;
    /**
     * - The timeout option name
     *    users can increase the timeout.
     */
    timeoutCapName?: string | undefined;
    /**
     * - Set to true in order to allow test
     *    packages installation.
     */
    allowTestPackages?: boolean | undefined;
    /**
     * - Set to true to install the app on sdcard
     *    instead of the device memory.
     */
    useSdcard?: boolean | undefined;
    /**
     * - Set to true in order to grant all the
     *    permissions requested in the application's manifest
     *    automatically after the installation is completed
     *    under Android 6+.
     */
    grantPermissions?: boolean | undefined;
    /**
     * - Set it to false if you don't want
     *    the application to be upgraded/reinstalled
     *    if it is already present on the device.
     */
    replace?: boolean | undefined;
    /**
     * - Forcefully disables incremental installs if set to `true`.
     *    Read https://developer.android.com/preview/features#incremental
     *    for more details.
     */
    noIncremental?: boolean | undefined;
};
export type InstallOrUpgradeOptions = {
    /**
     * - The count of milliseconds to wait until the
     *    app is installed.
     */
    timeout?: number | undefined;
    /**
     * - Set to true in order to allow test
     *    packages installation.
     */
    allowTestPackages?: boolean | undefined;
    /**
     * - Set to true to install the app on SDCard
     *    instead of the device memory.
     */
    useSdcard?: boolean | undefined;
    /**
     * - Set to true in order to grant all the
     *    permissions requested in the application's manifest
     *    automatically after the installation is completed
     *    under Android 6+.
     */
    grantPermissions?: boolean | undefined;
    /**
     * - Set to `true` in order to always prefer
     *    the current build over any installed packages having
     *    the same identifier
     */
    enforceCurrentBuild?: boolean | undefined;
};
export type InstallOrUpgradeResult = {
    /**
     * - Equals to `true` if the target app has been uninstalled
     *   before being installed
     */
    wasUninstalled: boolean;
    /**
     * - One of `adb.APP_INSTALL_STATE` states, which reflects
     * the state of the application before being installed.
     */
    appState: InstallState;
};
export type ApkStrings = {
    /**
     * parsed resource file
     * represented as JSON object
     */
    apkStrings: import("@appium/types").StringRecord;
    /**
     * the path to the extracted file on the local file system
     */
    localPath?: string | undefined;
};
export type AppInfo = {
    /**
     * - Package name, for example 'com.acme.app'.
     */
    name: string;
    /**
     * - Version code.
     */
    versionCode?: number | null | undefined;
    /**
     * - Version name, for example '1.0'.
     */
    versionName?: string | null | undefined;
    /**
     * - true if the app is installed on the device under test.
     */
    isInstalled?: boolean | null | undefined;
};
export type ApkUtils = typeof apkUtilsMethods;
export const REMOTE_CACHE_ROOT: "/data/local/tmp/appium_cache";
declare namespace apkUtilsMethods {
    let APP_INSTALL_STATE: Record<string, InstallState>;
    /**
     * @typedef {Object} IsAppInstalledOptions
     * @property {string} [user] - The user id
     */
    /**
     * Check whether the particular package is present on the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the package to check.
     * @param {IsAppInstalledOptions} [opts={}]
     * @return {Promise<boolean>} True if the package is installed.
     */
    function isAppInstalled(this: import("../adb.js").ADB, pkg: string, opts?: IsAppInstalledOptions | undefined): Promise<boolean>;
    /**
     * @typedef {Object} StartUriOptions
     * @property {boolean} [waitForLaunch=true] - if `false` then adb won't wait
     * for the started activity to return the control
     */
    /**
     * Start the particular URI on the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} uri - The name of URI to start.
     * @param {string} pkg - The name of the package to start the URI with.
     * @param {StartUriOptions} [opts={}]
     */
    function startUri(this: import("../adb.js").ADB, uri: string, pkg: string, opts?: StartUriOptions | undefined): Promise<void>;
    /**
     * @typedef {Object} StartAppOptions
     * @property {string} pkg - The name of the application package
     * @property {string} [activity] - The name of the main application activity.
     * This or action is required in order to be able to launch an app.
     * @property {string} [action] - The name of the intent action that will launch the required app.
     * This or activity is required in order to be able to launch an app.
     * @property {boolean} [retry=true] - If this property is set to `true`
     * and the activity name does not start with '.' then the method
     * will try to add the missing dot and start the activity once more
     * if the first startup try fails.
     * @property {boolean} [stopApp=true] - Set it to `true` in order to forcefully
     * stop the activity if it is already running.
     * @property {string} [waitPkg] - The name of the package to wait to on
     * startup (this only makes sense if this name is different from the one, which is set as `pkg`)
     * @property {string} [waitActivity] - The name of the activity to wait to on
     * startup (this only makes sense if this name is different from the one, which is set as `activity`)
     * @property {number} [waitDuration] - The number of milliseconds to wait until the
     * `waitActivity` is focused
     * @property {string|number} [user] - The number of the user profile to start
     * the given activity with. The default OS user profile (usually zero) is used
     * when this property is unset
     * @property {boolean} [waitForLaunch=true] - if `false` then adb won't wait
     * for the started activity to return the control
     * @property {string} [category]
     * @property {string} [flags]
     * @property {string} [optionalIntentArguments]
     */
    /**
     * Start the particular package/activity on the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @param {StartAppOptions} startAppOptions - Startup options mapping.
     * @return {Promise<string>} The output of the corresponding adb command.
     * @throws {Error} If there is an error while executing the activity
     */
    function startApp(this: import("../adb.js").ADB, startAppOptions: StartAppOptions): Promise<string>;
    /**
     * Helper method to call `adb dumpsys window windows/displays`
     * @this {import('../adb.js').ADB}
     * @returns {Promise<string>}
     */
    function dumpWindows(this: import("../adb.js").ADB): Promise<string>;
    /**
     * @typedef {Object} PackageActivityInfo
     * @property {string?} appPackage - The name of application package,
     *                                  for example 'com.acme.app'.
     * @property {string?} appActivity - The name of main application activity.
     */
    /**
     * Get the name of currently focused package and activity.
     *
     * @this {import('../adb.js').ADB}
     * @return {Promise<PackageActivityInfo>} The mapping, where property names are 'appPackage' and 'appActivity'.
     * @throws {Error} If there is an error while parsing the data.
     */
    function getFocusedPackageAndActivity(this: import("../adb.js").ADB): Promise<PackageActivityInfo>;
    /**
     * Wait for the given activity to be focused/non-focused.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the package to wait for.
     * @param {string} activity - The name of the activity, belonging to that package,
     *                            to wait for.
     * @param {boolean} waitForStop - Whether to wait until the activity is focused (true)
     *                                or is not focused (false).
     * @param {number} [waitMs=20000] - Number of milliseconds to wait before timeout occurs.
     * @throws {error} If timeout happens.
     */
    function waitForActivityOrNot(this: import("../adb.js").ADB, pkg: string, activity: string, waitForStop: boolean, waitMs?: number | undefined): Promise<void>;
    /**
     * Wait for the given activity to be focused
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the package to wait for.
     * @param {string} act - The name of the activity, belonging to that package,
     *                            to wait for.
     * @param {number} [waitMs=20000] - Number of milliseconds to wait before timeout occurs.
     * @throws {error} If timeout happens.
     */
    function waitForActivity(this: import("../adb.js").ADB, pkg: string, act: string, waitMs?: number | undefined): Promise<void>;
    /**
     * Wait for the given activity to be non-focused.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the package to wait for.
     * @param {string} act - The name of the activity, belonging to that package,
     *                            to wait for.
     * @param {number} [waitMs=20000] - Number of milliseconds to wait before timeout occurs.
     * @throws {error} If timeout happens.
     */
    function waitForNotActivity(this: import("../adb.js").ADB, pkg: string, act: string, waitMs?: number | undefined): Promise<void>;
    /**
     * @typedef {Object} UninstallOptions
     * @property {number} [timeout] - The count of milliseconds to wait until the
     *                                      app is uninstalled.
     * @property {boolean} [keepData] - Set to true in order to keep the
     *                                        application data and cache folders after uninstall.
     * @property {boolean} [skipInstallCheck] Whether to check if the app is installed prior to
     * uninstalling it. By default this is checked.
     */
    /**
     * Uninstall the given package from the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the package to be uninstalled.
     * @param {UninstallOptions} [options={}] - The set of uninstall options.
     * @return {Promise<boolean>} True if the package was found on the device and
     *                   successfully uninstalled.
     */
    function uninstallApk(this: import("../adb.js").ADB, pkg: string, options?: UninstallOptions | undefined): Promise<boolean>;
    /**
     * Install the package after it was pushed to the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} apkPathOnDevice - The full path to the package on the device file system.
     * @param {import('./system-calls.js').ShellExecOptions} [opts={}] Additional exec options.
     * @throws {error} If there was a failure during application install.
     */
    function installFromDevicePath(this: import("../adb.js").ADB, apkPathOnDevice: string, opts?: import("./system-calls.js").ShellExecOptions | undefined): Promise<void>;
    /**
     * @typedef {Object} CachingOptions
     * @property {number} [timeout] - The count of milliseconds to wait until the
     * app is uploaded to the remote location.
     */
    /**
     * Caches the given APK at a remote location to speed up further APK deployments.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} apkPath - Full path to the apk on the local FS
     * @param {CachingOptions} [options={}] - Caching options
     * @returns {Promise<string>} - Full path to the cached apk on the remote file system
     * @throws {Error} if there was a failure while caching the app
     */
    function cacheApk(this: import("../adb.js").ADB, apkPath: string, options?: CachingOptions | undefined): Promise<string>;
    /**
     * @typedef {Object} InstallOptions
     * @property {number} [timeout=60000] - The count of milliseconds to wait until the
     *                                      app is installed.
     * @property {string} [timeoutCapName=androidInstallTimeout] - The timeout option name
     *                                                             users can increase the timeout.
     * @property {boolean} [allowTestPackages=false] - Set to true in order to allow test
     *                                                 packages installation.
     * @property {boolean} [useSdcard=false] - Set to true to install the app on sdcard
     *                                         instead of the device memory.
     * @property {boolean} [grantPermissions=false] - Set to true in order to grant all the
     *                                                permissions requested in the application's manifest
     *                                                automatically after the installation is completed
     *                                                under Android 6+.
     * @property {boolean} [replace=true] - Set it to false if you don't want
     *                                      the application to be upgraded/reinstalled
     *                                      if it is already present on the device.
     * @property {boolean} [noIncremental=false] - Forcefully disables incremental installs if set to `true`.
     *                                             Read https://developer.android.com/preview/features#incremental
     *                                             for more details.
     */
    /**
     * Install the package from the local file system.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appPath - The full path to the local package.
     * @param {InstallOptions} [options={}] - The set of installation options.
     * @throws {Error} If an unexpected error happens during install.
     */
    function install(this: import("../adb.js").ADB, appPath: string, options?: InstallOptions | undefined): Promise<void>;
    /**
     * Retrieves the current installation state of the particular application
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appPath - Full path to the application
     * @param {string?} [pkg=null] - Package identifier. If omitted then the script will
     * try to extract it on its own
     * @returns {Promise<InstallState>} One of `APP_INSTALL_STATE` constants
     */
    function getApplicationInstallState(this: import("../adb.js").ADB, appPath: string, pkg?: string | null | undefined): Promise<InstallState>;
    /**
     * @typedef {Object} InstallOrUpgradeOptions
     * @property {number} [timeout=60000] - The count of milliseconds to wait until the
     *                                      app is installed.
     * @property {boolean} [allowTestPackages=false] - Set to true in order to allow test
     *                                                 packages installation.
     * @property {boolean} [useSdcard=false] - Set to true to install the app on SDCard
     *                                         instead of the device memory.
     * @property {boolean} [grantPermissions=false] - Set to true in order to grant all the
     *                                                permissions requested in the application's manifest
     *                                                automatically after the installation is completed
     *                                                under Android 6+.
     * @property {boolean} [enforceCurrentBuild=false] - Set to `true` in order to always prefer
     *                                                   the current build over any installed packages having
     *                                                   the same identifier
     */
    /**
     * @typedef {Object} InstallOrUpgradeResult
     * @property {boolean} wasUninstalled - Equals to `true` if the target app has been uninstalled
     *                                      before being installed
     * @property {InstallState} appState - One of `adb.APP_INSTALL_STATE` states, which reflects
     * the state of the application before being installed.
     */
    /**
     * Install the package from the local file system or upgrade it if an older
     * version of the same package is already installed.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appPath - The full path to the local package.
     * @param {string?} [pkg=null] - The name of the installed package. The method will
     * perform faster if it is set.
     * @param {InstallOrUpgradeOptions} [options={}] - Set of install options.
     * @throws {Error} If an unexpected error happens during install.
     * @returns {Promise<InstallOrUpgradeResult>}
     */
    function installOrUpgrade(this: import("../adb.js").ADB, appPath: string, pkg?: string | null | undefined, options?: InstallOrUpgradeOptions | undefined): Promise<InstallOrUpgradeResult>;
    /**
     * @typedef {Object} ApkStrings
     * @property {import('@appium/types').StringRecord} apkStrings parsed resource file
     * represented as JSON object
     * @property {string} [localPath] the path to the extracted file on the local file system
     */
    /**
     * Extract string resources from the given package on local file system.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appPath - The full path to the .apk(s) package.
     * @param {string?} [language=null] - The name of the language to extract the resources for.
     * The default language is used if this equals to `null`
     * @param {string?} [outRoot=null] - The name of the destination folder on the local file system to
     * store the extracted file to. If not provided then the `localPath` property in the returned object
     * will be undefined.
     * @return {Promise<ApkStrings>}
     */
    function extractStringsFromApk(this: import("../adb.js").ADB, appPath: string, language?: string | null | undefined, outRoot?: string | null | undefined): Promise<ApkStrings>;
    /**
     * Get the language name of the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @return {Promise<string>} The name of device language.
     */
    function getDeviceLanguage(this: import("../adb.js").ADB): Promise<string>;
    /**
     * Get the country name of the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @return {Promise<string>} The name of device country.
     */
    function getDeviceCountry(this: import("../adb.js").ADB): Promise<string>;
    /**
     * Get the locale name of the device under test.
     *
     * @this {import('../adb.js').ADB}
     * @return {Promise<string>} The name of device locale.
     */
    function getDeviceLocale(this: import("../adb.js").ADB): Promise<string>;
    /**
     * Make sure current device locale is expected or not.
     *
     * @this {import('../adb.js').ADB}
     * @privateRemarks FIXME: language or country is required
     * @param {string} [language] - Language. The language field is case insensitive, but Locale always canonicalizes to lower case.
     * @param {string} [country] - Country. The language field is case insensitive, but Locale always canonicalizes to lower case.
     * @param {string} [script] - Script. The script field is case insensitive but Locale always canonicalizes to title case.
     *
     * @return {Promise<boolean>} If current locale is language and country as arguments, return true.
     */
    function ensureCurrentLocale(this: import("../adb.js").ADB, language?: string | undefined, country?: string | undefined, script?: string | undefined): Promise<boolean>;
    /**
     * @typedef {Object} AppInfo
     * @property {string} name - Package name, for example 'com.acme.app'.
     * @property {number?} [versionCode] - Version code.
     * @property {string?} [versionName] - Version name, for example '1.0'.
     * @property {boolean?} [isInstalled] - true if the app is installed on the device under test.
     */
    /**
     * Get the package info from local apk file.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appPath - The full path to existing .apk(s) package on the local
     *                           file system.
     * @return {Promise<AppInfo|{}>} The parsed application information.
     */
    function getApkInfo(this: import("../adb.js").ADB, appPath: string): Promise<AppInfo | {}>;
    /**
     * Get the package info from the installed application.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg - The name of the installed package.
     * @return {Promise<AppInfo>} The parsed application information.
     */
    function getPackageInfo(this: import("../adb.js").ADB, pkg: string): Promise<AppInfo>;
    /**
     * Fetches base.apk of the given package to the local file system
     *
     * @this {import('../adb.js').ADB}
     * @param {string} pkg The package identifier (must be already installed on the device)
     * @param {string} tmpDir The destination folder path
     * @returns {Promise<string>} Full path to the downloaded file
     * @throws {Error} If there was an error while fetching the .apk
     */
    function pullApk(this: import("../adb.js").ADB, pkg: string, tmpDir: string): Promise<string>;
    /**
     * Activates the given application or launches it if necessary.
     * The action literally simulates
     * clicking the corresponding application icon on the dashboard.
     *
     * @this {import('../adb.js').ADB}
     * @param {string} appId - Application package identifier
     * @throws {Error} If the app cannot be activated
     */
    function activateApp(this: import("../adb.js").ADB, appId: string): Promise<void>;
}
//# sourceMappingURL=apk-utils.d.ts.map